(defpackage "WATT"
  (:use "COMMON-LISP")
  (:export "WATT"))

(in-package "WATT")

(defun dbg (&rest args)
  (apply #'format *trace-output* args)
  (finish-output *trace-output*))

#+(or)
(defun dbg (&rest args)
  (declare (ignore args))
  nil)

(declaim (type (integer 1 1000) *width* *height*)
         (integer *money*))

(defvar *world* nil)
(defvar *width* 1)
(defvar *height* 1)
(defvar *bases* nil)
(defvar *id* nil)
(defvar *capacity* nil)
(defvar *money* 0)
(defvar *robots* nil)
(defvar *packets* nil)
(defvar *we* nil)
(defvar *brain* nil)

(defstruct brain
  (path nil)
  (destination nil)
  (desire 'get-packet))

(defstruct base
  (i 1 :type (integer (1 1000)))
  (j 1 :type (integer (1 1000))))

(defstruct (field (:print-function print-field))
  (i 1 :type (integer 0 1001))
  (j 1 :type (integer 0 1001))
  (material nil)
  (content nil)
  (moves nil)#+(or)
  (targets (init-targets)))

(defun print-field (field &optional (stream t) depth)
  (declare (ignore depth))
  (print-unreadable-object (field stream :type t)
    (format stream "~A ~A MATERIAL ~A CONTENT ~A"
            (field-i field)
            (field-j field)
            (field-material field)
            (field-content field))))

(defstruct robot
  (id 0 :type integer)
  (i 1 :type (integer (1 1000)))
  (j 1 :type (integer (1 1000)))
  (packets nil))

(defstruct packet
  (id 0 :type integer)
  (weight nil :type (or null integer))
  (robot nil)
  (bad nil)
  (i 1 :type (integer (1 1000)))
  (j 1 :type (integer (1 1000)))
  (dest-i nil :type (or null (integer (1 1000))))
  (dest-j nil :type (or null (integer (1 1000)))))

(declaim (inline get-robot))

(defun get-robot (id)
  (gethash id *robots*))

(defun get-packet (id)
  (gethash id *packets*))

(defun del-packet (id)
  (remhash id *packets*))

(defsetf get-robot (id) (robot)
  `(setf (gethash ,id *robots*) ,robot))

(defsetf get-packet (id) (packet)
  `(setf (gethash ,id *packets*) ,packet))

(defmacro dopackets ((packet) &body body)
  (let ((id (gensym)))
    `(block nil
       (maphash (lambda (,id ,packet)
                  (declare (ignore ,id))
                  ,@body)
                *packets*))))

(defun robot-move (id direction)
#+(or)  (dbg "~&robot-move ~A ~A" id direction)
  (let ((robot (get-robot id))
        (vec (ecase direction
               (north '(0 1))
               (south '(0 -1))
               (east '(1 0))
               (west '(-1 0)))))
    (unless robot
      (error "Robot moving ~A doesn't exist." direction))
    (let ((field (field (robot-i robot) (robot-j robot))))
      (setf (field-content field)
            (delete robot (field-content field))))
    (incf (robot-i robot) (car vec))
    (incf (robot-j robot) (cadr vec))
    (push robot (field-content (field (robot-i robot) (robot-j robot))))))

(defun get-make-packet (id)
  (let ((packet (get-packet id)))
    (or packet
        (setf (get-packet id) (make-packet :id id)))))

(defun robot-pickup (id packet-id)
#+(or)  (dbg "~&robot-pickup ~A ~A" id packet-id)
  (let ((robot (get-robot id))
        (packet (get-make-packet packet-id)))
    (unless robot
      (error "Unknown robot ~A picked up packet ~A" id packet-id))
    (setf (robot-packets robot) (nconc (robot-packets robot) (list packet)))
    (setf (packet-robot packet) robot)
    (let ((field (field (robot-i robot) (robot-j robot))))
      (setf (field-content field)
            (delete packet (field-content field))))))

(defun robot-drop (id packet-id)
#+(or)  (dbg "~&robot-drop ~A ~A" id packet-id)
  (let ((robot (get-robot id))
        (packet (get-packet packet-id)))
    (unless robot
      (error "Unknown robot ~A dropped packet ~A" id packet-id))
    (unless packet
      (setq packet (make-packet :id packet-id))
      (setf (get-packet packet-id) packet))
    (setf (robot-packets robot) (delete packet (robot-packets robot)
                                        :test #'eq))
    (setf (packet-robot packet) nil)
    (setf (packet-i packet) (robot-i robot))
    (setf (packet-j packet) (robot-j robot))
    (let ((field (field (robot-i robot) (robot-j robot))))
      (if (and (packet-dest-i packet)
               (eq field (field (packet-dest-i packet)
                                (packet-dest-j packet))))
          (del-packet packet-id)
        (push packet (field-content field))))))

(defun robot-appear (id i j)
#+(or)  (dbg "~&robot-appear ~A ~A ~A" id i j)
  (unless (and (<= 1 i *width*)
               (<= 1 j *height*))
    (error "Robot appearing at illegal coordinates (~A,~A)." i j))
  (let ((robot (get-robot id)))
    (if robot
        (let* ((old-i (robot-i robot))
               (old-j (robot-j robot))
               (field (field old-i old-j)))
          (setf (field-content field)
                (delete robot (field-content field)))
          (setf (robot-i robot) i
                (robot-j robot) j))
      (setq robot (make-robot :id id :i i :j j)))
    (when (= id *id*)
      (setq *we* robot))
    (let ((field (field i j)))
      (push robot (field-content field))
      (setf (get-robot id) robot))))

#+(or)
(defun init-targets ()
  (make-hash-table :test #'eq))
#+(or)
(declaim (inline get-target))
#+(or)
(defun get-target (dest field)
  (gethash dest (field-targets field)))
#+(or)
(defsetf get-target (dest targets) (way)
  `(setf (gethash ,dest (field-targets ,targets)) ,way))

(defun has-base (field)
  (find-if #'base-p (field-content field)))

(defun has-mark (field)
  (find 'mark (field-content field)))

(defun mark-field (field)
  (push 'mark (field-content field)))

(defun show-world (&optional (world *world*))
  (loop for j from *height* downto 1 do
        (fresh-line)
        (loop for i from 1 to *width* do
              (let ((field (aref world i j)))
                (write-char (case (field-material field)
                              (wall #\#)
                              (water #\~)
                              (otherwise (cond ((has-mark field) #\*)
                                               ((has-base field) #\@)
                                               (t #\.))))))))
  (values))

(defun make-world (width height)
  (setq *robots* (make-hash-table))
  (setq *packets* (make-hash-table))
  (setq *width* width)
  (setq *height* height)
  (setq *bases* nil)
  (setq *we* nil)
  (setq *brain* (make-brain))
  (incf width)
  (incf height)
  (let ((ret (make-array (list (1+ width) (1+ height)))))
    (loop for i to width do
          (loop for j to height do
                (setf (aref ret i j)
                      (if (or (zerop i) (zerop j)
                              (= i width) (= j height))
                          (make-field :material 'wall :i i :j j)
                        (make-field :i i :j j)))))
    (setq *world* ret)))

(declaim (inline field))

(defun field (i j)
  (declare (type (simple-array t (* *)) *world*)
           (type (integer 0 1001) i j))
  (aref *world* i j))

(defun skip-ws ()
  (let ((char (read-char nil nil nil)))
    (case char
      ((#\Space #\Tab) (skip-ws))
      ((nil) nil)
      (otherwise (unread-char char)))))

(defun read-int ()
  (skip-ws)
  (flet ((read-sign ()
           (let ((next (read-char)))
             (cond ((char= next #\-) -1)
                   (t (unread-char next) 1)))))
    (let ((sign (read-sign))
          (ret 0))
      (let ((peek (peek-char nil nil nil nil)))
        (unless peek
          (error "Unexpected EOF after - sign."))
        (unless (digit-char-p peek)
          (error "Unexpected character ~:C in integer." peek)))
      (loop for char = (read-char nil nil nil) while char do
            (let ((dig (digit-char-p char)))
              (cond (dig (setq ret (+ dig (* 10 ret))))
                    (t (unread-char char) (return)))))
      (* sign ret))))

(defun read-dimensions ()
  (prog1
      (list (read-int) (read-int))
    (skip-nl)))

(defun skip-nl ()
  (skip-ws)
  (let ((char (read-char nil nil nil)))
    (unless (eql char #\Newline)
      (error "Unexpected character ~:C before newline." char))))

(defun read-world ()
  (loop for j from 1 to *height* do
        (loop for i from 1 to *width* do
              (let ((char (read-char nil nil nil))
                    (field (field i j)))
                (case char
                  (#\. nil)
                  (#\~ (setf (field-material field) 'water))
                  (#\# (setf (field-material field) 'wall))
                  (#\@ (let ((base (make-base :i i :j j)))
                         (push base *bases*)
                         (push base (field-content field))))
                  ((nil) (error "Unexpected EOF in board."))
                  (otherwise (error "Unexpected character ~:C in board." char)))))
        (skip-nl)))

(defun read-config ()
  (setq *id* (read-int)
        *capacity* (read-int)
        *money* (read-int))
  (skip-nl))

(defun read-response ()
  (skip-ws)
  (labels ((read-sym ()
             (skip-ws)
             (let ((char (read-char nil nil nil)))
               (case char
                 (#\N 'north)
                 (#\S 'south)
                 (#\E 'east)
                 (#\W 'west)
                 (#\P 'pickup)
                 (#\D 'drop)
                 (#\X 'x)
                 (#\Y 'y)
                 (#\Newline 'eol)
                 (#\# 'hash)
                 ((nil) (error "Unexpected EOF in server response."))
                 (otherwise
                  (error "Unexpected character ~:C in action." char)))))
           (read-action (id)
             (let ((sym (read-sym)))
               (ecase sym
                 ((north south east west) (robot-move id sym))
                 (pickup (robot-pickup id (read-int)))
                 (drop (robot-drop id (read-int)))
                 (x (let ((x (read-int))
                          (sym (read-sym)))
                      (if (eq sym 'y)
                          (robot-appear id x (read-int))
                        (error "Unexpected token ~A in location." sym))))
                 (hash (return-from read-action
                         (read-action (read-int)))) ; empty action list :-(
                 (eol (return-from read-action nil)))
               (skip-ws)
               (let ((char (read-char nil nil nil)))
                 (case char
                   (#\Newline nil)
                   (#\# (unread-char char) (read-response))
                   (otherwise (unread-char char) (read-action id)))))))
    (let ((char (read-char nil nil nil)))
      (when (and char (not (char= char #\Newline)))
        (if (char= char #\#)
            (read-action (read-int))
          (error "Unexpected character ~:C in server response." char))))))


(defun read-init ()
  (format t "Player~%")
  (finish-output)
  (apply #'make-world (read-dimensions))
  (read-world)
  (read-config)
  (read-response))

(defun show-robots ()
  (maphash (lambda (id robot)
             (format *trace-output* "~&Robot ~A: ~A" id robot))
           *robots*))

(defun show-packets ()
  (maphash (lambda (id packet)
             (format *trace-output* "~&Packet ~A: ~A" id packet))
           *packets*))

(defun test-init ()
  (let ((some-config "7 5
..@....
.......
##.~~~~
...~~~~
.......
3 25 1000
#2 X 2 Y 1 #1 X 1 Y 2 #3 X 1 Y 1
#2 P 17 #1 P 12
#1 D 12 #2 W #3 N #3 P 12
17 133 28 50 89 11 57 78
"))
    (with-input-from-string (*standard-input* some-config)
      (read-init)
      (read-response)
      (read-response)
      (read-packet-list)
      (show-robots)
      (show-packets))))

(define-condition game-over ()
  ())

(defun read-packet-list (&optional (we *we*) acc)
  (skip-ws)
  (let ((char (read-char nil nil nil)))
    (cond ((null char) (error 'game-over))
          ((char= char #\Newline) acc)
          ((or (char= char #\-) (digit-char-p char))
           (unread-char char)
           (let* ((id (read-int))
                  (i (read-int))
                  (j (read-int))
                  (weight (read-int))
                  (packet (get-make-packet id)))
             (cond ((and (<= 1 i *width*)
                         (<= 1 j *height*)
                         (not (field-material (field i j))))
                    (setf (packet-dest-i packet) i)
                    (setf (packet-dest-j packet) j))
                   (t (setf (packet-bad packet) t)))
             (setf (packet-weight packet) weight)
             (when (> weight *capacity*)
               (setf (packet-bad packet) t))
             (setf (packet-i packet) (robot-i we))
             (setf (packet-j packet) (robot-j we))
             (setf (packet-robot packet) nil)
             (read-packet-list we (cons packet acc))))
          (t (error "Unexpected character ~:C in packet list." char)))))

(defconstant +vecs+ '((1 . 0) (0 . 1) (-1 . 0) (0 . -1)))

(defun moves (i j)
  (let ((field (field i j)))
    (or (field-moves field)
        (loop with ret = nil
              for (dx . dy) in +vecs+ do
              (let* ((next-i (+ i dx))
                     (next-j (+ j dy))
                     (field (field next-i next-j)))
                (unless (field-material field)
                  (push field ret)))
              finally (return (setf (field-moves field) ret))))))

(defstruct path
  (field)
  (prev nil)
  (len 0 :type integer)
  (total 0 :type integer))

(defun field-distance (field1 field2)
  (let ((i1 (field-i field1))
        (i2 (field-i field2))
        (j1 (field-j field1))
        (j2 (field-j field2)))
    (+ (abs (- i1 i2))
       (abs (- j1 j2)))))

(defun coord-distance (i1 j1 i2 j2)
  (+ (abs (- i1 i2))
     (abs (- j1 j2))))

;;; Inspired by an algorithm in Peter Norvig's "Paradigms of AI Programming"

(defun search-path (paths target garbage)
  (flet ((find-path-in (field list)
           (find field list :key #'path-field :test #'eq))
         (better-path (p1 p2)
           (< (path-total p1) (path-total p2)))
         (insert-path (new list)
           (merge 'list (list new) list #'< :key #'path-total)))
    (declare (inline find-path-in better-path insert-path))
    (cond ((null paths) nil)
          ((eq target (path-field (car paths)))
           (car paths))
          (t (let* ((path (pop paths))
                    (field (path-field path)))
               (setf (gethash (path-field path) garbage) path)
               (dolist (move (moves (field-i field) (field-j field)))
                 (let* ((len (1+ (path-len path)))
                        (diff (field-distance move target))
                        (new-path (make-path :field move
                                             :prev path
                                             :len len
                                             :total (+ len diff)))
                        (tmp nil))
                   (cond ((setq tmp (find-path-in move paths))
                          (when (better-path new-path tmp)
                            (setq paths (insert-path new-path
                                                     (delete tmp paths)))))
                         ((setq tmp (gethash move garbage))
                          (when (better-path new-path tmp)
                            (setq paths (insert-path new-path paths))
                            (remhash (path-field tmp) garbage)))
                         (t (setq paths (insert-path new-path paths)))))))
             (search-path paths target garbage)))))

(defun show-path (path)
  (loop for head = path then (path-prev head) while head do
        (pprint (path-field head))))

(defun rev-path (path)
  (loop with ret = nil
        for x = path then (path-prev x)
        while x do
        (push (path-field x) ret)
        finally (return ret)))

(defun find-path (from to)
  (let* ((garbage (make-hash-table :test #'eq))
         (path (search-path (list (make-path :field from
                                             :total (field-distance from
                                                                    to)))
                            to garbage)))
    (clrhash garbage)
    (values (cdr (rev-path path)) (and path (path-len path)))))

(defun play-game ()
  (read-init)
  (loop
   (issue-command (delete-if #'packet-bad (read-packet-list)))
   (finish-output)
   (read-response)))

(defun connect-to-server (&key (server #+(or)"icfp1.cse.ogi.edu"
                                       "localhost") (port 20005 #+(or)22007))
  (with-open-stream (stream (comm:open-tcp-stream server port))
    (let ((*standard-input* stream)
          (*standard-output* stream))
      (play-game))))

(defun log-server (&key (server #+(or) "icfp1.cse.ogi.edu"
                                "localhost") (port 20005 #+(or) 22007))
  (with-open-stream (stream (comm:open-tcp-stream server port))
    (let ((*standard-output* (make-broadcast-stream *trace-output* stream))
          (*standard-input* (make-echo-stream stream *trace-output*)))
      (play-game))))

(defun send-command (bid &rest args)
  (format t "~D " bid)
  (decf *money* (abs bid))
  (apply #'format t args))

(defun set-desire (desire)
  (setf (brain-desire *brain*) desire))

(defvar *last-goto* nil)

(defun goto-field (field bid)
  (let ((i (robot-i *we*))
        (j (robot-j *we*))
        (new-i (field-i field))
        (new-j (field-j field)))
    (when (and (eq *last-goto* field)
               (> *money* (1+ bid)))
      (incf bid))
    (setq *last-goto* field)
    (cond ((= i new-i)
           (ecase (- new-j j)
             (1 (send-command bid "Move N~%"))
             (-1 (send-command bid "Move S~%"))))
          ((= j new-j)
           (ecase (- new-i i)
             (1 (send-command bid "Move E~%"))
             (-1 (send-command bid "Move W~%"))))
          (t (error "Illegal move.")))))

(defun find-nearest-packet ()
  (let ((here (field (robot-i *we*) (robot-j *we*)))
        (nearest nil)
        (dist nil)
        (path nil))
    (dolist (base *bases*)
      (multiple-value-bind (way len)
          (find-path here (field (base-i base) (base-j base)))
        (when (and len (or (null nearest) (< len dist)))
          (setq nearest (field (base-i base) (base-j base))
                dist len
                path way))))
    (dopackets (packet)
      (unless (or (packet-robot packet) (packet-bad packet))
        (multiple-value-bind (way len)
            (find-path here (field (packet-i packet) (packet-j packet)))
          (when (and len (or (null nearest) (< len dist)))
            (setq nearest (field (packet-i packet) (packet-j packet))
                  dist len
                  path way)))))
    (values path nearest)))

(defun find-pickable (packets)
  (let* ((ours (robot-packets *we*))
         (here (field (robot-i *we*) (robot-j *we*)))
         (room (- *capacity* (loop for i in ours sum (packet-weight i)))))
    (loop with ret = nil
          for i in packets while (plusp room) do
          (when (<= (packet-weight i) room)
            (let ((len (nth-value 1
                                  (find-path here (field (packet-dest-i i)
                                                         (packet-dest-j i))))))
              (cond ((null len)
                     (setf (packet-bad i) t))
                    (t (push (packet-id i) ret)
                       (decf room (packet-weight i))))))
          finally (return ret))))

(defun pick-some-packets (packets)
  (send-command 1 "Pick~{ ~D~}~%"
                (find-pickable packets)))

(defun no-packets-here (here)
  (let ((junk nil))
    (dopackets (packet)
      (when (eq here (field (packet-i packet) (packet-j packet)))
        (push packet junk)))
    (dolist (packet junk)
      (del-packet packet))))

(defun goto-packet (packets)
  (let ((tmp nil)
        (here (field (robot-i *we*) (robot-j *we*))))
    (cond ((robot-packets *we*)
           (set-desire 'deliver)
           (deliver-packet packets))
          (packets (pick-some-packets packets))
          ((and (setq tmp (pop (brain-path *brain*)))
                (= 1 (field-distance tmp here)))
           (goto-field tmp 1))
          ((setq tmp (brain-destination *brain*))
           (cond ((eq tmp here)
                  (no-packets-here here)
                  (setf (brain-destination *brain*) nil)
                  (goto-packet packets))
                 (t (setf (brain-path *brain*) (find-path here tmp))
                    (goto-field (pop (brain-path *brain*)) 1))))
          (t (multiple-value-bind (path dest)
                 (find-nearest-packet)
               (cond ((null path) (idle))
                     (t (setf (brain-path *brain*) (cdr path))
                        (setf (brain-destination *brain*) dest)
                        (goto-field (car path) 1))))))))

(defun drop-some-packets ()
  (let* ((here (field (robot-i *we*) (robot-j *we*)))
         (drops (loop with ret = nil
                      for packet in (robot-packets *we*) do
                      (when (eq (field (packet-dest-i packet)
                                       (packet-dest-j packet))
                                here)
                        (push (packet-id packet) ret))
                      finally (return ret))))
    (send-command 1 "Drop~{ ~D~}~%" drops)))

(defun deliver-packet (packets)
  (let* ((packet (car (robot-packets *we*)))
         (i (robot-i *we*))
         (j (robot-j *we*))
         (here (field i j))
         (tmp nil)
         (bdest (brain-destination *brain*)))
    (cond (packet (let* ((dest-i (packet-dest-i packet))
                         (dest-j (packet-dest-j packet))
                         (dest (field dest-i dest-j)))
                    (cond ((eq here dest) (drop-some-packets))
                          ((and (brain-path *brain*) (eq dest bdest)
                                (= 1 (field-distance (car (brain-path *brain*))
                                                     here)))
                           (if (and packets
                                    (setq tmp (find-pickable packets)))
                               (send-command 1 "Pick~{ ~D~}~%" tmp)
                             (goto-field (pop (brain-path *brain*)) 1)))
                          (t (setf (brain-path *brain*) (find-path here dest)
                                   (brain-destination *brain*) dest)
                             (goto-field (pop (brain-path *brain*)) 1)))))
          (t (set-desire 'get-packet)
             (setf (brain-path *brain*) nil
                   (brain-destination *brain*) nil)
             (goto-packet packets)))))

(defun idle ()
  (let* ((fields (moves (robot-i *we*) (robot-j *we*)))
         (len (length fields)))
    (goto-field (nth (random len) fields) 1)))

(defun issue-command (packets)
  (if (null packets)
      (let* ((here (field (robot-i *we*) (robot-j *we*)))
             (base (has-base here)))
        (when base
          (setf (field-content here) (delete base (field-content here)
                                             :test #'eq))
          (setf *bases* (delete base *bases* :test #'eq))))
    (setq packets (sort packets #'> :key #'packet-weight)))
  (ecase (brain-desire *brain*)
    (get-packet (goto-packet packets))
    (deliver (deliver-packet packets))))

(defun watt ()
  (let* ((args system:*line-arguments-list*)
         (server (second args))
         (port (third args)))
    (handler-case
        (setq port (parse-integer port))
      (error nil (format *error-output* "Usage: ~A <SERVER> <PORT>~%"
                         (or (car args) "watt"))
             (lw:quit :status 1)))
    (handler-case
        (connect-to-server :server server :port port)
      (error nil nil))))
